#include "tap.h"

#define POWEROFF_ID 0x8beef00c
TAP_ID(POWEROFF_ID);
TAP_PROGRAM_NAME( "PowerOff" );
TAP_AUTHOR_NAME("DeadBeef");
TAP_DESCRIPTION("Signals power-off event to TAPs");
TAP_ETCINFO(__DATE__);

#define JAL_CMD 0x0c000000
#define JR_CMD 0x03e00008
#define JAL_MASK 0xff000000
#define LUI_A2_CMD 0x3c060000
#define OR_A2_CMD 0x34c60000
#define JUMP_CMD 0x08000000
#define JAL_CMD 0x0c000000
#define NOP_CMD 0x00000000
#define REL_ADDR(x) (((dword)(x) & 0x03ffffff) >> 2)
#define ABS_ADDR(x) (0x80000000 | (((dword)(x) & 0x03ffffff) << 2))

typedef word (*CheckRecording)(int number);
typedef word (*StartRecording)(int number);
typedef word (*GenerateEvent)(int event, int param1, int param2);

typedef struct
{
  CheckRecording checkRecording;
  StartRecording startRecording;
  GenerateEvent generateEvent;
  bool bStarted;
} HookData;

dword *pFwSendToDisplay = NULL;
static HookData hookData;

//-----------------------------------------------------------------------------
// fwHook() evaluates the strings sent to the display. If the string
// contains "OFF" then the function initiates the power off procedure.
// Since this function is called from the FW context the following is
// Neither calling TAP API functions nor accessing global data is
// allowed.
// Background info: Even though TF5500 and TF5000MP have different displays
// the firmware contains the function for the 7-segment display which is
// called during the shutdown just before evaluation of recording.
void fwHook(char mask, char *pString, HookData *pHookData)
{
  // check for the "OFF" string
  if((pString[0] != 0x3f) || (pString[1] != 0x71) || (pString[2] != 0x71))
  {
    // no match, return to proceed with normal operation
    return;
  }

  // "OFF" string matched
  if(!(pHookData->checkRecording(0) || pHookData->checkRecording(1)))
  {
    // no recording is currently active, start dummy recording to
    // prevent an immediate shutdown 
    pHookData->bStarted = pHookData->startRecording(0);
  }

  // generate a power-off event (as opposed to the API function the
  // first parameter is 0xe300)
  pHookData->generateEvent(0xe300, RKEY_Power, 0);
}

//-----------------------------------------------------------------------------
// exitHook() is supposed to execute the two instructions replaced by the
// installed hook and continue with the execution of the firmware code at the
// position right after where the CPU jumped to the hook.
// exitHook() is actually part of the fwHook() function. It has been separated
// for convinience reasons (i.e. now it is easier to replace the instructions
// without performing a search).
void exitHook()
{
  __asm__ __volatile__ (
  "nop\n"           /* two placeholders for the replaced FW instruction */
  "nop\n"           /*  */
  "nop\n"           /* one jump separator */
  "nop\n"           /* one placeholder for the j(ump) instruction (back to the FW code) */
  "nop\n");         /* one jump separator */
}

//-----------------------------------------------------------------------------
// enterHook() is a proxy function which passes the call from the FW to the
// fwHook function. The function to be called shall take three parameters ($a0 = mask, 
// $a1 = pointer to FW display data, $a2 = pointer to hook parameters).
static void enterHook(char mask, char *pString)
{
  __asm__ __volatile__ (
  "lui     $06,0x0\n"       /* This register is used for the address of the structure containing hook parameters */
  "or      $06,$06,0x0\n"   /* the two instructions are to be replaced with the actual address on start-up */
  "add     $29,$29,-20\n"   /* allocate stack space */
  "sw      $31,0($29)\n"    /* push return address to stack */
  "sw      $04,4($29)\n"    /* push $a0 to stack, because it might be modified in the called function */
  "sw      $05,8($29)\n"    /* push $a1 to stack, because it might be modified in the called function */
  "bal     fwHook\n"        /* call the function */
  "lw      $05,8($29)\n"    /* restore $a1 from stack */
  "lw      $04,4($29)\n"    /* restore $a0 from stack */
  "lw      $31,0($29)\n"    /* restore return address from stack */
  "add     $29,$29,20\n"    /* free stack space */
  "b       exitHook\n");    /* jump to the exit code segment */
}

//-----------------------------------------------------------------------------
// definitions for the find routine
#define ADD_SP_MASK 0xffffff00
#define ADD_SP_CMD 0x27bdff00
#define CMD_MASK 0xfc00ffff
#define AND1_CMD 0x30000001
#define AND2_CMD 0x30000002
#define AND4_CMD 0x30000004
#define AND8_CMD 0x30000008
#define AND10_CMD 0x30000010
#define ADD_CMD 0x24000003
#define ADD_FLAG 0x80000000
#define MAX_FUNC_LENGTH 100

//-----------------------------------------------------------------------------
// findDisplayFunction() retrieves the address of the display function
// sending data to the 7-segment display (even if it is not physically
// availbale). 
dword findDisplayFunction(dword *pData, int size)
{
  dword rc = 0;
  int i = 0, start = 0;

  while(i < size)
  {
    if((pData[i] & ADD_SP_MASK) == ADD_SP_CMD)
    {
      // detected start of a function, reset variables
      start = i;
      rc = 0;
    }
    else
    {
      // search for typical attributes of the display function
      unsigned int tmp = pData[i] & CMD_MASK;
      switch(tmp)
      {
      case AND1_CMD:
      case AND2_CMD:
      case AND4_CMD:
      case AND8_CMD:
      case AND10_CMD:
        rc |= tmp;
        break;
      case ADD_CMD:
        rc |= ADD_FLAG;
        break;
      }

      if(rc == (AND1_CMD | AND2_CMD | AND4_CMD | AND8_CMD | AND10_CMD | ADD_FLAG))
      {
        // return if all attributes matched
        return (dword)(pData + start);
      }

      if(i > (start + MAX_FUNC_LENGTH))
      {
        // segment too long, reset
        rc = 0;
      }

    }

    i++;
  }

  return 0;
}

//-----------------------------------------------------------------------------
// findCheckRecording() retrieves the address of the function checking whether
// recording is active.
// The firmware-internal check function is the first function called in
// TAP_Hdd_StopRecord(). So find the call and convert the address.
dword findCheckRecording()
{
  dword *pSrc = (dword*)TAP_Hdd_StopRecord;
  int i = 0;

  while(pSrc[i] != JR_CMD)
  {
    if((pSrc[i] & JAL_MASK) == JAL_CMD)
    {
      // this is the wrapped call to the actual FW implementation
      // return the absolute address encoded in the opcode
      return ABS_ADDR(pSrc[i]);
    }

    i++;
  }

  return 0;
}

//-----------------------------------------------------------------------------
// findStartRecording() retrieves the address of the function starting a record.
// The firmware-internal start function is the first function called in
// TAP_Hdd_StartRecord(). So find the call and convert the address.
dword findStartRecording()
{
  dword *pSrc = (dword*)TAP_Hdd_StartRecord;
  int i = 0;

  while(pSrc[i] != JR_CMD)
  {
    if((pSrc[i] & JAL_MASK) == JAL_CMD)
    {
      // this is the wrapped call to the actual FW implementation
      // return the absolute address encoded in the opcode
      return ABS_ADDR(pSrc[i]);
    }

    i++;
  }

  return 0;
}

//-----------------------------------------------------------------------------
// findGenerateEvent() retrieves the address of the FW function generating
// events. The firmware-internal function is the first function called in
// TAP_GenerateEvent(). So find the call and convert the address.
dword findGenerateEvent()
{
  dword *pSrc = (dword*)TAP_GenerateEvent;
  int i = 0;

  while(pSrc[i] != JR_CMD)
  {
    if((pSrc[i] & JAL_MASK) == JAL_CMD)
    {
      // this is the wrapped call to the actual FW implementation
      // return the absolute address encoded in the opcode
      return ABS_ADDR(pSrc[i]);
    }

    i++;
  }

  return 0;
}

//-----------------------------------------------------------------------------
void uninstallHook()
{
  dword *pSrc = (dword*)exitHook;

  pFwSendToDisplay[0] = pSrc[0];
  pFwSendToDisplay[1] = pSrc[1];
}

//-----------------------------------------------------------------------------
dword TAP_EventHandler( word event, dword param1, dword param2 )
{
  static dword tick = 0;

  if((event == EVT_KEY))
  {
    switch(param1)
    {
    case  RKEY_Power:
      // got the power-off event
      // start timer
      tick = TAP_GetTick();
      TAP_Print("OFF event (%d, %d)\n", hookData.bStarted, tick);
      break;
    case RKEY_1:
      // got an exit event
      TAP_Print("TAP terminated\n");
      uninstallHook();
      TAP_Exit();
      param1 = 0;
    }
  }
  else if(event == EVT_IDLE)
  {
    // check whether a dummy recording is active
    if((tick != 0) && hookData.bStarted)
    {
      // check the timer (5 seconds)
      if((tick + 500) < TAP_GetTick())
      {
        TAP_Print("stopping dummy record (%d)\n", TAP_GetTick());
        // stop the dummy recording, it must always be in slot 0
        // because it is only started if recording was not active
        TAP_Hdd_StopRecord(0);
        hookData.bStarted = FALSE;
      }
    }
  }

  return param1;
}

//-----------------------------------------------------------------------------
int TAP_Main(void)
{
  dword *pBuffer = NULL;

  TAP_Print("Starting TAP\n");

  // initialize the global data
  hookData.checkRecording = (CheckRecording)findCheckRecording();
  hookData.startRecording = (StartRecording)findStartRecording();
  hookData.generateEvent = (GenerateEvent)findGenerateEvent();
  hookData.bStarted = FALSE;

  // September 13, 2005
  //pFwSendToDisplay = (dword*)0x8000654c;

  // uncomment if the TF5000Display TAP is not installed
  pFwSendToDisplay = (dword*)findDisplayFunction((dword*)0x80000000, 0x10000);

  if(hookData.checkRecording == NULL)
  {
    TAP_Print("checkRecording not found\n");
    return 0;
  }

  if(hookData.startRecording == NULL)
  {
    TAP_Print("startRecording not found\n");
    return 0;
  }

  if(hookData.generateEvent == NULL)
  {
    TAP_Print("generateEvent not found\n");
    return 0;
  }

  if(pFwSendToDisplay == NULL)
  {
    TAP_Print("sendToDisplay not found\n");
    return 0;
  }

  TAP_Print("checkRecording = 0x%08x\n", hookData.checkRecording);
  TAP_Print("startRecording = 0x%08x\n", hookData.startRecording);
  TAP_Print("generateEvent = 0x%08x\n", hookData.generateEvent);
  TAP_Print("sendToDisplay = 0x%08x\n", pFwSendToDisplay);

  // initialize exitHook()
  pBuffer = (dword*)exitHook;
  // save the original FW instructions in exitHook() which will be replaced below
  pBuffer[0] = pFwSendToDisplay[0];
  pBuffer[1] = pFwSendToDisplay[1];
  pBuffer[2] = NOP_CMD;
  // set the return address to the original code
  pBuffer[3] = JUMP_CMD | REL_ADDR(pFwSendToDisplay + 2);
  pBuffer[4] = NOP_CMD;

  // initialize enterHook() with the target address
  // validate the expected command order
  pBuffer = (dword*)enterHook;
  if((pBuffer[0] != LUI_A2_CMD) || (pBuffer[1] != OR_A2_CMD))
  {
    TAP_Print("hook initialization failed\n");
    return 0;
  }

  // initialize memory address
  pBuffer[0] = LUI_A2_CMD | (((dword)&hookData >> 16) & 0xffff);
  pBuffer[1] = OR_A2_CMD | ((dword)&hookData & 0xffff);

  TAP_Print("Hook initialized (0x%x)\n", enterHook);

  // install the hook
  pFwSendToDisplay[0] = JUMP_CMD | REL_ADDR(enterHook);
  pFwSendToDisplay[1] = NOP_CMD;

  TAP_Print("Hook installed\n");

  // go to TSR mode
	return 1;
}

	
